/*
 * Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 * Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package edu.ucsc.its.temerity.shared.ui.elements

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.animation.AnimatedVisibilityScope
import androidx.compose.animation.core.animateFloatAsState
import androidx.compose.animation.fadeIn
import androidx.compose.animation.fadeOut
import androidx.compose.animation.slideInHorizontally
import androidx.compose.animation.slideOutHorizontally
import androidx.compose.animation.slideOutVertically
import androidx.compose.foundation.Image
import androidx.compose.foundation.ScrollbarStyle
import androidx.compose.foundation.background
import androidx.compose.foundation.border
import androidx.compose.foundation.gestures.Orientation
import androidx.compose.foundation.gestures.draggable
import androidx.compose.foundation.gestures.rememberDraggableState
import androidx.compose.foundation.gestures.scrollBy
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.PaddingValues
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.Spacer
import androidx.compose.foundation.layout.fillMaxHeight
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.height
import androidx.compose.foundation.layout.heightIn
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.width
import androidx.compose.foundation.layout.widthIn
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.LazyItemScope
import androidx.compose.foundation.lazy.LazyListScope
import androidx.compose.foundation.lazy.items
import androidx.compose.foundation.lazy.rememberLazyListState
import androidx.compose.foundation.shape.RoundedCornerShape
import androidx.compose.foundation.text.BasicText
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.rounded.ChevronRight
import androidx.compose.material3.CardDefaults
import androidx.compose.material3.HorizontalDivider
import androidx.compose.material3.Icon
import androidx.compose.material3.IconButton
import androidx.compose.material3.OutlinedCard
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.rememberCoroutineScope
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.draw.rotate
import androidx.compose.ui.geometry.Size
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.PathFillType
import androidx.compose.ui.graphics.SolidColor
import androidx.compose.ui.graphics.StrokeCap
import androidx.compose.ui.graphics.StrokeJoin
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.ui.graphics.vector.path
import androidx.compose.ui.layout.onSizeChanged
import androidx.compose.ui.text.TextStyle
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.text.style.TextOverflow
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.toSize
import androidx.navigation.NavGraph
import androidx.navigation.NavHostController
import androidx.navigation.compose.NavHost
import com.composables.composetheme.ComposeTheme
import com.composables.composetheme.material3.colorScheme
import com.composables.composetheme.material3.typography
import com.composables.core.Menu
import com.composables.core.MenuButton
import com.composables.core.MenuContent
import com.composables.core.MenuItem
import com.composables.core.rememberMenuState
import kotlinx.coroutines.launch

/**
 * A composable function that creates a box with an adaptive aspect ratio.
 * The box maintains a specified aspect ratio until it reaches a maximum height,
 * beyond which the width can grow indefinitely while the height remains constant.
 *
 * @param modifier The [Modifier] to be applied to the box.
 * @param maxHeight The maximum height of the box in pixels. Once this height is reached, the width can grow indefinitely.
 * @param aspectRatio The aspect ratio (width / height) to maintain until the maximum height is reached.
 * @param content The content to be displayed inside the box. This is a composable lambda.
 */
@Composable
fun AdaptiveAspectRatioBox(
  modifier: Modifier = Modifier,
  maxHeight: Float = 250f,
  aspectRatio: Float = 16f / 9f,
  content: @Composable () -> Unit,
) {
  var size by remember { mutableStateOf(Size.Zero) }
  val maxWidth = if (size.height >= maxHeight) Float.POSITIVE_INFINITY else maxHeight * aspectRatio

  Box(
    modifier = modifier
      .onSizeChanged {
        size = it.toSize()
      }
      .widthIn(max = maxWidth.dp)
      .heightIn(max = maxHeight.dp),
  ) {
    content()
  }
}

@Composable
fun ScrollableContent(modifier: Modifier = Modifier, content: LazyListScope.() -> Unit) {
  val unHoverColor = ComposeTheme.colorScheme.surfaceContainerHighest
  val hoverColor = ComposeTheme.colorScheme.secondaryContainer
  val scrollbarStyle = ScrollbarStyle(
    minimalHeight = 32.dp,
    thickness = 12.dp,
    shape = RoundedCornerShape(8.dp),
    hoverDurationMillis = 160,
    unhoverColor = unHoverColor,
    hoverColor = hoverColor,
  )

  val scrollState = rememberLazyListState()
  val coroutineScope = rememberCoroutineScope()

  Box(modifier = modifier.fillMaxSize()) {
    LazyColumn(
      modifier = Modifier
        .draggable(
          orientation = Orientation.Horizontal,
          state = rememberDraggableState { delta ->
            coroutineScope.launch {
              scrollState.scrollBy(-delta)
            }
          },
        )
        .padding(horizontal = 4.dp),
      state = scrollState,
      verticalArrangement = Arrangement.spacedBy(12.dp),
      contentPadding = PaddingValues(horizontal = 24.dp, vertical = 24.dp),
    ) {
      content()
    }
    VerticalScrollbar(
      adapter = rememberScrollbarAdapter(scrollState),
      modifier = Modifier
        .align(Alignment.CenterEnd)
        .fillMaxHeight()
        .padding(MARGIN_SCROLLBAR),
      style = scrollbarStyle,
    )
  }
}

@Composable
fun <T> ScrollableList(items: List<T> = emptyList(), itemLayout: @Composable LazyItemScope.(T) -> Unit) {
  val scrollableListItems = remember { mutableStateOf(items) }
  ScrollableContent {
    items(scrollableListItems.value) { item ->
      itemLayout(item)
    }
  }
}

@Composable
fun CardRow(
  title: String,
  subtitle: String,
  modifier: Modifier = Modifier,
  content: @Composable AnimatedVisibilityScope.() -> Unit,
) {
  val textColor = ComposeTheme.colorScheme.primary
  val spacerColor = ComposeTheme.colorScheme.background

  OutlinedCard {
    var expanded by remember { mutableStateOf(false) }
    val degrees by animateFloatAsState(if (expanded) -90f else 90f)

    Column(
      modifier = modifier
        .fillMaxWidth()
        .widthIn(min = 200.dp, max = 400.dp),
    ) {
      AdaptiveAspectRatioBox(
        modifier = Modifier
          .clip(CardDefaults.outlinedShape),
      ) {
        Gradient()
      }
      Column(Modifier.padding(start = 16.dp, end = 16.dp)) {
        Row(verticalAlignment = Alignment.CenterVertically) {
          Text(
            text = title,
            color = textColor,
            style = ComposeTheme.typography.titleLarge,
            modifier = Modifier.weight(1f),
            maxLines = 1,
            overflow = TextOverflow.Ellipsis,
          )
          IconButton(onClick = { expanded = expanded.not() }) {
            Icon(
              imageVector = Icons.Rounded.ChevronRight,
              contentDescription = if (expanded) "Hide details" else "Show more details",
              modifier = Modifier.rotate(degrees),
            )
          }
        }
        Row {
          Text(
            text = subtitle,
            color = textColor,
          )
          Spacer(
            modifier = Modifier
              .width(8.dp)
              .background(spacerColor),
          )
        }
        Spacer(
          modifier = Modifier
            .height(16.dp)
            .background(spacerColor),
        )
        Box {
          this@Column.AnimatedVisibility(visible = expanded) {
            content()
          }
          this@Column.AnimatedVisibility(visible = expanded) {
            HorizontalDivider()
          }
        }
      }
    }
  }
}

@Composable
fun NewNavHost(navController: NavHostController, navGraph: NavGraph) = NavHost(
  navController,
  navGraph,
  enterTransition = {
    fadeIn() + slideInHorizontally(initialOffsetX = { it / 2 })
  },
  exitTransition = {
    fadeOut() + slideOutHorizontally(targetOffsetX = { it / 2 })
  },
  popExitTransition = {
    fadeOut() + slideOutVertically(targetOffsetY = { -it })
  },
)

@Composable
fun chevronDown(): ImageVector {
  val iconColor = ComposeTheme.colorScheme.onSurface
  return remember {
    ImageVector.Builder(
      name = "ChevronDown",
      defaultWidth = 16.dp,
      defaultHeight = 16.dp,
      viewportWidth = 24f,
      viewportHeight = 24f,
    ).apply {
      path(
        fill = null,
        fillAlpha = 1.0f,
        stroke = SolidColor(iconColor),
        strokeAlpha = 1.0f,
        strokeLineWidth = 2f,
        strokeLineCap = StrokeCap.Round,
        strokeLineJoin = StrokeJoin.Round,
        strokeLineMiter = 1.0f,
        pathFillType = PathFillType.NonZero,
      ) {
        moveTo(6f, 9f)
        lineToRelative(6f, 6f)
        lineToRelative(6f, -6f)
      }
    }.build()
  }
}

@Composable
fun DropdownMenu(
  items: List<String>,
  onSelectItem: (String) -> Unit,
  modifier: Modifier = Modifier,
) {
  val state = rememberMenuState(expanded = false)
  var selectedItem by remember { mutableStateOf("Select") }
  val surfaceColor = ComposeTheme.colorScheme.surface
  val textColor = ComposeTheme.colorScheme.onSurface

  Box(modifier = modifier) {
    Menu(modifier = Modifier.align(Alignment.TopStart).width(240.dp), state = state) {
      val degrees by animateFloatAsState(if (state.expanded) -180f else 0f)

      MenuButton(
        Modifier.clip(RoundedCornerShape(6.dp)).background(surfaceColor)
          .border(1.dp, Color(0xFFBDBDBD), RoundedCornerShape(6.dp)),
      ) {
        Row(
          verticalAlignment = Alignment.CenterVertically,
          modifier = Modifier.padding(horizontal = 14.dp, vertical = 10.dp),
        ) {
          BasicText(selectedItem, modifier = Modifier.widthIn(min = 160.dp), style = TextStyle(fontWeight = FontWeight(500), color = textColor))
          Spacer(Modifier.width(4.dp))
          Image(chevronDown(), null, Modifier.rotate(degrees))
        }
      }

      MenuContent(
        modifier = Modifier.padding(top = 4.dp).width(320.dp).clip(RoundedCornerShape(6.dp))
          .border(1.dp, Color(0xFFE0E0E0), RoundedCornerShape(6.dp)).background(surfaceColor).padding(4.dp),
        exit = fadeOut(),
      ) {
        items.forEach {
          MenuItem(modifier = Modifier.clip(RoundedCornerShape(6.dp)), onClick = {
            selectedItem = it
            onSelectItem(it)
          }) {
            BasicText(it, Modifier.fillMaxWidth().padding(vertical = 10.dp, horizontal = 10.dp), style = TextStyle(color = textColor))
          }
        }
      }
    }
  }
}
