/*
 * Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 * Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package edu.ucsc.its.temerity.shared.viewmodel

import androidx.compose.runtime.Composable
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.setValue
import edu.ucsc.its.temerity.model.User
import edu.ucsc.its.temerity.shared.viewmodel.UserEditEvent.ClearField
import edu.ucsc.its.temerity.shared.viewmodel.UserEditEvent.LoadUser
import edu.ucsc.its.temerity.shared.viewmodel.UserEditEvent.Save
import edu.ucsc.its.temerity.shared.viewmodel.UserEditEvent.UpdateField
import kotlinx.coroutines.flow.Flow

sealed class UserEditEvent {
  data class LoadUser(val user: User?) : UserEditEvent()
  data object Save : UserEditEvent()
  data class ClearField(val fieldName: String) : UserEditEvent()
  data class UpdateField(val fieldName: String, val value: String) : UserEditEvent()
}

data class UserEditState(
  var firstName: String = "",
  var lastName: String = "",
  var loginId: String = "",
  var emailAddress: String = "",
  val timezone: String = "",
  val role: String = "",
  val lockUser: Boolean = false,
)

class UserEditViewModel : MoleculeViewModel<UserEditEvent, UserEditState>() {
  @Composable
  override fun models(events: Flow<UserEditEvent>): UserEditState = userEditPresenter(events)

  @Composable
  fun userEditPresenter(events: Flow<UserEditEvent>): UserEditState {
    var state by remember { mutableStateOf(UserEditState()) }

    LaunchedEffect(Unit) {
      events.collect {
        when (it) {
          is LoadUser -> {
            val user = it.user
            if (user != null) {
              state = state.copy(
                firstName = user.firstName,
                lastName = user.lastName,
                loginId = user.loginId,
                emailAddress = user.emailAddress,
                timezone = user.timezone,
                role = user.userType,
              )
            }
          }
          is Save -> {
            // Save the user edits
          }
          is ClearField -> {
            state = updateField(it.fieldName, "", state)
          }

          is UpdateField -> {
            state = updateField(it.fieldName, it.value, state)
          }
        }
      }
    }
    return state
  }

  private fun updateField(fieldName: String, value: String, state: UserEditState): UserEditState = when (fieldName) {
    "firstName" -> state.copy(firstName = value)
    "lastName" -> state.copy(lastName = value)
    "loginId" -> state.copy(loginId = value)
    "emailAddress" -> state.copy(emailAddress = value)
    "timezone" -> state.copy(timezone = value)
    "role" -> state.copy(role = value)
    else -> state
  }

  suspend fun loadUser(user: User?) {
    events.emit(LoadUser(user))
  }

  suspend fun clearField(fieldName: String) {
    events.emit(ClearField(fieldName))
  }

  suspend fun updateField(fieldName: String, value: String) {
    events.emit(UpdateField(fieldName, value))
  }

  suspend fun saveUserEdits() {
    events.emit(Save)
  }
}
