/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
import org.jetbrains.kotlin.gradle.ExperimentalKotlinGradlePluginApi
import io.github.andreabrighi.gradle.gitsemver.conventionalcommit.ConventionalCommit
import org.jetbrains.dokka.gradle.engine.parameters.VisibilityModifier.Public

plugins {
    alias(libs.plugins.kotlinMultiplatform)
    alias(libs.plugins.androidLibrary)
    alias(libs.plugins.ksp)
    alias(libs.plugins.ktorfit)
    alias(libs.plugins.kotlinxSerialization)
    alias(libs.plugins.kotestMultiplatform)

    id("convention.formattingLib")
    id("convention.version")
    alias(libs.plugins.dokka)
    id("convention.publication")

    alias(libs.plugins.dataframe)

}

buildscript {
    dependencies {
        classpath(libs.dokka.base)
        classpath(libs.build.conventionalCommits)
    }
}

gitSemVer {
    commitNameBasedUpdateStrategy(ConventionalCommit::semanticVersionUpdate)
}

dependencies {
    dokkaPlugin(libs.build.androidDocumentationPlugin)
}

kotlin {
    jvmToolchain(libs.versions.java.get().toInt())
    applyDefaultHierarchyTemplate()
    explicitApi()

    @OptIn(ExperimentalKotlinGradlePluginApi::class)
    compilerOptions {
        freeCompilerArgs.add("-Xexpect-actual-classes")
    }

    jvm()
    androidTarget()

    sourceSets {
        all {
            languageSettings.apply {
                optIn("kotlin.RequiresOptIn")
            }
        }
        val commonMain by getting {
            dependencies {
                api(libs.ktor.client.core)
                implementation(libs.bundles.ktor)
                implementation(libs.bundles.sandwich)
                api(libs.kotlinx.datetime)
                implementation(libs.koin.core)
                implementation(libs.ktorfit.lib)
                implementation(libs.kotlinx.coroutines.core)
                implementation(libs.kermit)
                implementation(libs.cache4k)
                implementation(libs.kmpIo)
            }
        }
        val commonTest by getting {
            dependencies {
                implementation(libs.kotest.framework.engine)
                implementation(libs.kotest.assertions.core)
                implementation(libs.kotest.framework.datatest)
                implementation(libs.ktor.test)
                implementation(libs.konsist)
            }
        }
        val jvmMain by getting {
            dependencies {
                implementation(libs.ktor.client.java)
                implementation(libs.slf4j)
            }
        }
        val androidMain by getting {
            dependencies {
                implementation(libs.ktor.client.android)
            }
        }
        val jvmTest by getting {
            dependencies {
                implementation(libs.dotenv.vault)

                implementation(libs.kotest.runner.junit5)
                implementation(libs.kotlinx.coroutines.slf4j)
                implementation(libs.kotlinx.coroutines.debug)
                implementation(libs.kotlinx.dataframe)
                // Required for report exports
                // Used to convert dataframe rows to serializable objects
                implementation(libs.klaxon)

                implementation(libs.koin.test)
                implementation(libs.koin.test.junit5)
            }
        }
    }
}

tasks.withType<Test> {
    useJUnitPlatform()
}

tasks.named("sourcesJar") {
    dependsOn("kspCommonMainKotlinMetadata")
}

android {
    compileSdk = libs.versions.compileSdk.get().toInt()
    defaultConfig {
        minSdk = libs.versions.minSdk.get().toInt()
        testInstrumentationRunner = "androidx.test.runner.AndroidJUnitRunner"
    }
    namespace = "edu.ucsc.its.${rootProject.name}"
}

dokka {
    moduleName.set(rootProject.name)
    dokkaSourceSets.commonMain {
        sourceLink {
            localDirectory.set(file("src/commonMain/kotlin"))
            remoteUrl("https://git.ucsc.edu/wnwalker/${rootProject.name}/-/tree/main/${rootProject.name}/src/commonMain/kotlin/edu/ucsc/its/${rootProject.name}")
            remoteLineSuffix.set("#L")
        }
        skipEmptyPackages.set(true)
        documentedVisibilities.set(setOf(Public))
    }
    pluginsConfiguration.html {
        footerMessage = "Developed by William Walker Copyright © 2024 Regents of the University of California"
    }
}

publishing {
    publications {
        withType<MavenPublication> {
            // Workaround for gradle issue: https://github.com/gradle/gradle/issues/26091
            // https://github.com/voize-gmbh/reakt-native-toolkit/commit/baf0392b65e3ab0305b491f64c45a3a5fc6c86a4
            val publication = this
            val dokkaJar = tasks.register<Jar>("${publication.name}DokkaJar") {
                group = JavaBasePlugin.DOCUMENTATION_GROUP
                description = "Assembles Kotlin docs with Dokka into a Javadoc jar"
                archiveClassifier.set("javadoc")
                from(tasks.named("dokkaGenerate"))
                // Each archive name should be distinct, to avoid implicit dependency issues.
                // We use the same format as the sources Jar tasks.
                // https://youtrack.jetbrains.com/issue/KT-46466
                archiveBaseName.set("${archiveBaseName.get()}-${publication.name}")
            }

            artifact(dokkaJar)

            // Version set by Gradle Git SemVer plugin: https://github.com/DanySK/git-sensitive-semantic-versioning-gradle-plugin
            group = "edu.ucsc.its"

            pom {
                name.set("Temerity Library")
                description.set("A Kotlin Multiplatform library for interacting with the YuJa EVP API")
                url.set("https://git.ucsc.edu/wnwalker/${rootProject.name}")

                licenses {
                    license {
                        name.set("GNU Lesser General Public License, Version 2.1")
                        url.set("https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt")
                    }
                }
                developers {
                    developer {
                        id.set("wnwalker")
                        name.set("William Walker")
                        email.set("wnwalker@ucsc.edu")
                        organizationUrl.set("https://its.ucsc.edu")
                    }
                }

                issueManagement {
                    system.set("GitLab")
                    url.set("https://git.ucsc.edu/wnwalker/${rootProject.name}/-/issues")
                }

                scm {
                    connection.set("scm:git:https://git.ucsc.edu/wnwalker/${rootProject.name}.git")
                    developerConnection.set("scm:git:ssh://git@git.ucsc.edu:wnwalker/${rootProject.name}.git")
                    url.set("https://git.ucsc.edu/wnwalker/${rootProject.name}")
                }
            }
        }
    }
}
