/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity

import edu.ucsc.its.temerity.core.TemDsl
import io.ktor.client.HttpClient
import io.ktor.client.HttpClientConfig
import io.ktor.client.engine.HttpClientEngineConfig
import io.ktor.client.engine.HttpClientEngineFactory
import io.ktor.client.plugins.logging.LoggingConfig
import kotlin.time.Duration

/**
 * Configuration for the Temerity client.
 * Adapted from the tmdb-api project: https://github.com/MoviebaseApp/tmdb-kotlin/raw/refs/heads/main/tmdb-api/src/commonMain/kotlin/app/moviebase/tmdb/TmdbClientConfig.kt
 *
 * @property serviceUrl Specifies URL of platform instance to make API calls to
 * @property serviceToken Specifies the authorization token to use for API calls
 * @property optDebugEnabled Specifies whether debug logging should be enabled
 *
 * @property expectSuccess Specifies whether to expect successful responses by default
 * @property useWebTimeout Specifies whether to use a timeout  other than default for web requests
 * @property webTimeout Specifies the timeout [Duration] to use when making web requests
 * @property cacheTimeout Specifies the timeout [Duration] to use when storing cache entries. Affects keep-alive time for cached instance data like user role type fields.
 */
@TemDsl
public class TemClientConfig {

  public var serviceUrl: String? = null
  public var serviceToken: String? = null
  public var optDebugEnabled: Boolean = false

  public var expectSuccess: Boolean = true
  public var useWebTimeout: Boolean = false

  @Suppress("MemberVisibilityCanBePrivate")
  public var webTimeout: Duration? = null

  // Default cache timeout is set to 15 minutes if this is left unspecified; @see [edu.ucsc.its.temerity.core.Temerity] init block
  public var cacheTimeout: Duration? = null

  /**
   * Configures the HttpClient with the provided block.
   * @param () -> HttpClient The block to configure the HttpClient.
   */
  internal var httpClientConfigBlock: (HttpClientConfig<*>.() -> Unit)? = null
  internal var httpClientBuilder: (() -> HttpClient)? = null
  internal var httpClientLoggingBlock: (LoggingConfig.() -> Unit)? = null

  public fun logging(block: LoggingConfig.() -> Unit) {
    httpClientLoggingBlock = block
  }

  /**
   * Set custom HttpClient configuration for the default HttpClient.
   */
  public fun httpClient(block: HttpClientConfig<*>.() -> Unit) {
    httpClientConfigBlock = block
  }

  /**
   * Creates an custom [HttpClient] with the specified [HttpClientEngineFactory] and optional [block] configuration.
   * Note that the Temerity config will be added afterwards.
   */
  public fun <T : HttpClientEngineConfig> httpClient(
    engineFactory: HttpClientEngineFactory<T>,
    block: HttpClientConfig<T>.() -> Unit = {},
  ) {
    httpClientBuilder = {
      HttpClient(engineFactory, block)
    }
  }

  public companion object {

    internal fun withToken(serviceToken: String): TemClientConfig = TemClientConfig().apply {
      this.serviceToken = serviceToken
    }
  }
}
