/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity

import edu.ucsc.its.temerity.model.AuditLogEntry
import edu.ucsc.its.temerity.model.Course
import edu.ucsc.its.temerity.model.Device
import edu.ucsc.its.temerity.model.DeviceRecordingSession
import edu.ucsc.its.temerity.model.EventType
import edu.ucsc.its.temerity.model.FolderPermissions
import edu.ucsc.its.temerity.model.Group
import edu.ucsc.its.temerity.model.NewUser
import edu.ucsc.its.temerity.model.User
import edu.ucsc.its.temerity.model.UserGroup
import edu.ucsc.its.temerity.model.UserRecordingSession
import edu.ucsc.its.temerity.model.UserUpdate
import io.ktor.client.statement.HttpResponse
import kotlinx.datetime.DateTimeUnit
import kotlinx.datetime.LocalDate
import kotlinx.datetime.LocalDateTime
import kotlinx.datetime.format.DateTimeFormat
import kotlinx.datetime.format.char
import kotlinx.datetime.plus

/**
 * This interface defines top-level library functions for interacting with the enterprise video platform API.
 * The functions defined in this interface are implemented by the [edu.ucsc.its.temerity.core.Temerity] class.
 * @see [edu.ucsc.its.temerity.core.Temerity]
 */

public interface TemerityApi {

  // User API public library functions

  /**
   * Fetches the list of users from the platform.
   * @return A list of User objects
   * @throws RuntimeException for:
   *  - Network issues
   *  Includes HttpResponse if the request fails due to:
   *  - Invalid service endpoint (404 not found)
   *  - Invalid service token (401 unauthorized)
   */
  public suspend fun getUsers(): List<User>

  /**
   * Fetches a specific user from the platform.
   * @param userId The ID of the user to fetch.
   * @return A list of User objects.
   * @throws RuntimeException for:
   *  - Network issues
   *  Includes HttpResponse if the request fails due to:
   *  - Invalid service endpoint (404 not found)
   *  - Invalid service token (401 unauthorized)
   */
  public suspend fun getUser(userId: Long): User

  /**
   * Creates a new user on the platform.
   * @param primaryIdentifier An optional primary identifier attribute for the user, such as an email address or login ID.
   * Used to uniquely identify the user in the platform when logging in from an external source, such as LMS.
   * @param newUser A NewUser object representing the details of the user to create, some of which are required.
   * @see <a href="https://support.yuja.com/hc/en-us/articles/360049580714-YuJa-API#h_01H8YCQKJ0GKB8J5TMZ3DFA4V7">Required new user fields</a>
   * @return HttpResponse object indicating success or failure of the request.
   * @throws RuntimeException for:
   *  - Network issues
   *  Includes HttpResponse if the request fails due to:
   *  - Invalid NewUser object (415 unsupported media type)
   *  - Invalid service endpoint (404 not found)
   *  - Invalid service token (401 unauthorized)
   */
  public suspend fun createUser(primaryIdentifier: String = "", newUser: NewUser): HttpResponse

  /**
   * Updates a specific user on the platform.
   * @param userId The unique platform ID of the user record to write.
   * @param userUpdate A UserUpdate object representing only the detail fields you wish to update of a specific user.
   * @see <a href="https://support.yuja.com/hc/en-us/articles/360049580714-YuJa-API#h_01H8YCQKJ0JRN8T6G7QNMHP15F">All optional user update fields</a>
   * @return HttpResponse object indicating success or failure of the request.
   * @throws RuntimeException for:
   *  - Network issues
   *  Includes HttpResponse if the request fails due to:
   *  - Invalid service endpoint (404 not found)
   *  - Invalid service token (401 unauthorized)
   */
  public suspend fun updateUser(userId: Long, userUpdate: UserUpdate): HttpResponse

  /**
   * Deletes a specific user from the platform.
   * @param userId The ID of the user to delete.
   * @return HttpResponse object indicating success or failure of the request:
   *  - 200 OK if the user was successfully deleted, with payload:
   *    "Deleted user: $userId success."
   *  - 404 Not Found if the user does not exist, with payload:
   *    "Error - No such user."
   * @throws RuntimeException for:
   *  - Network issues
   *  Includes HttpResponse if the request fails due to:
   *  - Invalid service endpoint (404 not found)
   *  - Invalid service token (401 unauthorized)
   */
  public suspend fun deleteUser(userId: Long): HttpResponse

  /**
   * Fetches a list of groups that a user is a member of.
   * @param userId The ID of the user to fetch the groups for.
   * @return A list of UserGroup objects.
   * @throws Exception if the groups cannot be fetched.
   */
  public suspend fun getUserGroups(userId: Long): List<UserGroup>

  /**
   * Fetches a list of groups that a user is the content owner of.
   * @param userId The ID of the user to fetch the groups for.
   * @return A list of UserGroup objects.
   * @throws Exception if the groups cannot be fetched.
   */
  public suspend fun getUserGroupsOwned(userId: Long): List<UserGroup>

  public suspend fun createUserWithExternalId(primaryIdentifier: String, newUser: NewUser): HttpResponse

  public suspend fun addFolderPermission(perms: FolderPermissions.Action): HttpResponse

  public suspend fun removeFolderPermission(perms: FolderPermissions.Action)

  public suspend fun getMediaMetadata(linkageValue: String)

  // Group API public library functions

  public suspend fun getGroups(
    // TODO: Make this field instance-wide and depend on API version
    paginated: Boolean = true,
  ): List<Group>

  public suspend fun getCourse(courseCode: String): Course

  public suspend fun getCourses(
    // TODO: Make this field instance-wide and depend on API version
    paginated: Boolean = true,
  ): List<Course>

  /**
   * Fetches a list of audit log entries from the platform for a given time period and list of event types.
   * @param startTime The start date for the log entries.
   * @param endTime The end date for the log entries.
   * @param eventTypeList A list of event types to filter the log entries.
   * @param sortOrder The order in which to sort the log entries.
   * @return A list of AuditLogEntry objects.
   */
  public suspend fun getAuditLogEntries(
    startTime: LocalDate,
    endTime: LocalDate,
    eventTypeList: List<EventType>? = null,
    sortOrder: AuditLogSortOrder? = null,
  ): List<AuditLogEntry>

  /**
   * Fetches a list of audit log entries from the platform for a given time period and a single event type.
   * @param startTime The start date for the log entries.
   * @param endTime The end date for the log entries.
   * @param eventType A single event type to filter for among the log entries.
   * @param sortOrder The order in which to sort the log entries.
   * @return A list of AuditLogEntry objects.
   */
  public suspend fun getAuditLogEntries(
    startTime: LocalDate,
    endTime: LocalDate,
    eventType: EventType,
    sortOrder: AuditLogSortOrder? = null,
    // TODO: Make this field instance-wide and depend on API version
    paginated: Boolean = true,
  ): List<AuditLogEntry>

  /**
   * Fetches a list of devices registered with the current instance from the platform.
   * @return A list of Device objects.
   */
  public suspend fun getDevices(): List<Device>

  /**
   * Fetches a specific device from the platform.
   * @param deviceId The ID of the device to fetch.
   * @return A Device object.
   */
  public suspend fun getDevice(deviceId: Long): Device

  /**
   * Fetches a list of sessions a user has scheduled for a given time period.
   * @param userId The ID of the user to fetch the sessions for.
   * @param startTime The start date for the sessions.
   * @param endTime The end date for the sessions.
   * @return A list of UserRecordingSession objects.
   */
  public suspend fun getUserSessions(userId: String, startTime: LocalDate, endTime: LocalDate): List<UserRecordingSession>

  /**
   * Fetches a list of sessions on a device's schedule for a given time period.
   * @param deviceId The ID of the device to fetch the sessions for.
   * @param startTime The start date for the sessions. Defaults to the current date.
   * @param endTime The end date for the sessions. Defaults to one week from the current date.
   * @return A list of DeviceRecordingSession objects.
   */
  public suspend fun getDeviceSchedule(deviceId: Long, startTime: LocalDate = currentDate(), endTime: LocalDate = currentDate().plus(1, DateTimeUnit.WEEK)): List<DeviceRecordingSession>

  /**
   * Fetches a file containing storage usage details of a specific group.
   * @param groupId The ID of the group to fetch the storage usage for.
   */
  public suspend fun getStorageAnalyticsReport(groupId: Long): ByteArray

  public suspend fun refreshCachedUserRoles(): List<String>

  public suspend fun getCachedUserRoles(refresh: Boolean): List<String>
}

// Objects used by Temerity implementations, and possibly library consumers

public val AUDIT_LOG_REQUEST_DATE_FORMAT: DateTimeFormat<LocalDate> = LocalDate.Format {
  dayOfMonth()
  char('/')
  monthNumber()
  char('/')
  year()
}
public val AUDIT_LOG_TIMESTAMP_FORMAT: DateTimeFormat<LocalDateTime> = LocalDateTime.Format {
  year()
  char('-')
  monthNumber()
  char('-')
  dayOfMonth()
  char(' ')
  hour()
  char(':')
  minute()
  char(':')
  second()
}
public val SCHEDULED_SESSION_DATE_FORMAT: DateTimeFormat<LocalDate> = LocalDate.Format {
  dayOfMonth()
  char('-')
  monthNumber()
  char('-')
  year()
}

/**
 * Enum class used to define the sort order for audit log entries.
 */
public enum class AuditLogSortOrder {
  NEW_FIRST,
  OLD_FIRST,
  ;

  public companion object {
    public fun fromBoolean(value: Boolean): AuditLogSortOrder = if (value) NEW_FIRST else OLD_FIRST
  }
}

/**
 * Sorts a list of [AuditLogEntry] objects by their creation date.
 * @param sortOrder The order in which to sort the entries.
 */
public fun ArrayList<AuditLogEntry>.sortByCreationDate(sortOrder: AuditLogSortOrder) {
  when (sortOrder) {
    AuditLogSortOrder.NEW_FIRST -> sortByDescending { auditLogEntry -> auditLogEntry.creationTimestamp }
    AuditLogSortOrder.OLD_FIRST -> sortBy { auditLogEntry -> auditLogEntry.creationTimestamp }
  }
}
