/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity.model

import edu.ucsc.its.temerity.extensions.LocalDateTimeSerializer
import kotlinx.datetime.LocalDateTime
import kotlinx.serialization.Serializable

/**
 * Enum class of YuJa API Audit Log event types
 *
 * This class is used to represent the different types of events that can be queried from the YuJa API's audit log endpoint.
 * @param value The string value of the event type as it is represented in the YuJa API. Used to make the API call.
 */
public enum class EventType(public val value: String) {
  ACTIVE_ARCHIVING_REQUEST("Active Archiving Request"),
  AD_ASTRA_SYNC_FAILED("Ad Astra Sync Failed"),
  ALL_CAPTIONS("All Captions"),
  ALL_USER_NOTIFICATIONS("All User Notifications"),
  ANONYMOUS_USER_QUIZ_ATTEMPT_STARTED("Anonymous User Quiz Attempt Started"),
  ANONYMOUS_USER_QUIZ_QUESTION_ANSWERED("Anonymous User Quiz Question Answered"),
  ANONYMOUS_USER_QUIZ_QUESTION_REACHED("Anonymous User Quiz Question Reached"),
  ANONYMOUS_USER_QUIZ_SUBMISSION("Anonymous User Quiz Submission"),
  AUTO_CAPTIONS("Auto-Captions"),
  AUTOMATED_SESSION_FAILED_TO_START("Automated Session failed to Start"),
  AUTOMATED_SESSION_MONITOR("Automated Session Monitor"),
  BLACKBOARD_COLLABORATE_IMPORTING("Blackboard Collaborate Importing"),
  BULK_BLACKBOARD_USER_IMPORT("Bulk Blackboard User Import"),
  BULK_WEBEX_USER_IMPORT("Bulk Webex User Import"),
  BULK_ZOOM_USER_IMPORT("Bulk Zoom User Import"),
  CAPTURE_ERROR("Capture Error"),
  CHANGE_OWNER("Change Owner"),
  CHANGE_THUMBNAIL("Change Thumbnail"),
  COMMENTS("Comments"),
  COURSES_AND_GROUPS("Courses and Groups"),
  CREATE_QUIZ("Create Quiz"),
  DELETE_CAPTIONS("Delete Captions"),
  DELETE_FOLDER("Delete Folder"),
  DELETE_MEDIA("Delete Media"),
  DELETE_QUIZ("Delete Quiz"),
  DELETE_TRANSCRIPT("Delete Transcript"),
  DELETE_USER("Delete User"),
  DEVICE_ERROR("Device Error"),
  DEVICE_FAILED_TO_START("Device Failed to Start"),
  DEVICE_STATUS_AND_ERROR_LOGS("Device Status And Error Logs"),
  DEVICE_STATUS("DEVICE_STATUS"),
  DOWNLOAD_CONTENT("Download Content"),
  DOWNLOAD_REQUEST("Download Request"),
  EDIT_MEDIA("Edit Media"),
  EDIT_SCHEDULE("Edit Schedule"),
  ERROR("ERROR"),
  Event("Event"),
  FAILED_LOG_IN("Failed Log In"),
  FAVORITE_MEDIA("Favorite Media"),
  FOLDER_MANAGEMENT("Folder Management"),
  GENERATE_PERMALINK("Generate Permalink"),
  GRADE_NOT_SENT_TO_LMS_GRADEBOOK("Grade Not Sent to LMS Gradebook"),
  HARDWARE_CAPTURE_HUB("Hardware Capture (HUB)"),
  HARDWARE_CAPTURE_HUB_CAPTURE_INITIATED("Hardware Capture (HUB) - Capture Initiated"),
  HDCP_SIGNAL_DETECTED("HDCP Signal Detected"),
  HIMALAYAS("Himalayas"),
  HIMALAYAS_ARCHIVING("Himalayas Archiving"),
  HIMALAYAS_RESTORE("Himalayas Restore"),
  HUB_CAPTURE_PROBLEM("Hub Capture Problem"),
  HUMAN_CAPTION_NEEDS_APPROVAL("Human Caption Needs Approval"),
  HUMAN_CAPTION_FINISHED_FROM_YUJA_PRO("Human-Caption Finished from Yuja Pro Captioning"),
  HUMAN_CAPTION_REJECTED_BY_YUJA_PRO("Human-Caption Rejected by Yuja Pro Captioning"),
  HUMAN_CAPTION_STARTED_FROM_YUJA_PRO("New Human-Caption Started from YuJa Pro Captioning"),
  HUMAN_CAPTIONS("Human-Captions"),
  IMPORTING_ZOOM_RECORDING("Importing Zoom Recording"),
  INTEGRATIONS("Integrations"),
  LMS_GRADEBOOK_FAILED_TO_SYNC("LMS Gradebook Failed to Sync"),
  LMS_GRADEBOOK_SUCCESSFULLY_SYNCED("LMS Gradebook Successfully Synced"),
  LMS_USER_QUIZ_ATTEMPT_STARTED("LMS User Quiz Attempt Started"),
  LMS_USER_QUIZ_QUESTION_ANSWERED("LMS User Quiz Question Answered"),
  LMS_USER_QUIZ_QUESTION_REACHED("LMS User Quiz Question Reached"),
  LMS_USER_QUIZ_SUBMISSION("LMS User Quiz Submission"),
  MEDIA_DOWNLOADS("Media Downloads"),
  MEDIA_MANAGEMENT("Media Management"),
  MEDIA_PLAYER_BRANDING("Media Player Branding"),
  MICROSOFT_TEAMS_IMPORTING("Microsoft Teams Importing"),
  MOBILE_DOWNLOADS("Mobile Downloads"),
  MOVE_FOLDER("Move Folder"),
  MOVE_MEDIA("Move Media"),
  NEW_ALERT("New Alert"),
  NEW_AUTO_CAPTION("New Auto-Caption"),
  NEW_BROWSER_CAPTURE("New Browser Capture"),
  NEW_CAPTURE("New Capture"),
  NEW_COURSE("New Course"),
  NEW_FOLDER("New Folder"),
  NEW_HUMAN_CAPTION_STARTED_FROM_YUJA_PRO("New Human-Caption Started from YuJa Pro Captioning"),
  NEW_LOG_IN("New Log In"),
  NEW_MEDIA("New Media"),
  NEW_MOBILE_LOG_IN("New Mobile Log In"),
  NEW_PUBLISH("New Publish"),
  NEW_SCHEDULE("New Schedule"),
  NEW_USER("New User"),
  NEW_VIDEO_CONFERENCE("New Video Conference"),
  NEW_VIEW("New View"),
  POST_QUIZ("Post Quiz"),
  PUBLISH_QUIZ("Publish Quiz"),
  QUIZ_GRADE_FOR_LMS_CALCULATED("Quiz Grade for LMS Calculated"),
  RECORDING_AND_STREAMING("Recording and Streaming"),
  RECORDING_ERROR("Recording Error"),
  RECURRING_ACTION_PERFORMED("Recurring Action Performed"),
  RECYCLE_FOLDER("Recycle Folder"),
  RECYCLE_MEDIA("Recycle Media"),
  RENAME_FOLDER("Rename Folder"),
  RENAME_MEDIA("Rename Media"),
  REPLACE_MEDIA("Replace Media"), // Not documented, but returned by API
  RESTORE_ACTIVE_ARCHIVE("Restore Active Archive"),
  RESTORE_FOLDER("Restore Folder"),
  RESTORE_MEDIA("Restore Media"),
  SAML_SSO_SETTINGS_UPDATED("SAML SSO Settings Updated"),
  SESSION("SESSION"),
  SHARE_FOLDER("Share Folder"),
  SHARE_MEDIA("Share Media"),
  SHARE_MEDIA_ASSET("Share Media Asset"),
  SOFTWARE_CAPTURE_MAC("Software Capture (Mac)"),
  SOFTWARE_CAPTURE_MAC_CAPTURE_INITIATED("Software Capture (Mac) - Capture Initiated"),
  SOFTWARE_CAPTURE_PC("Software Capture (PC)"),
  SOFTWARE_CAPTURE_PC_CAPTURE_INITIATED("Software Capture (PC) - Capture Initiated"),
  SUB_CLIPS("Sub-Clips"),
  UNDO_ACTIVE_ARCHIVE("Undo Active Archive"),
  UNFAVORITE_MEDIA("Unfavorite Media"),
  UNPUBLISH_FOLDER("Unpublish Folder"),
  UNPUBLISH_MEDIA("Unpublish Media"),
  UPLOAD_CAPTIONS("Upload Captions"),
  UPLOAD_TRANSCRIPT("Upload Transcript"),
  USER_CREATED_EMPTY_POLICY("User Created Empty Policy"),
  USER_DEMOTED("User Demoted"),
  USER_PROMOTED("User Promoted"),
  USER_QUIZ_QUESTION_ANSWERED("User Quiz Question Answered"),
  USER_QUIZ_QUESTION_REACHED("User Quiz Question Reached"),
  USER_QUIZ_SUBMISSION("User Quiz Submission"),
  USERS("Users"),
  VIDEO_CONFERENCE("Video Conference"),
  VIDEO_QUIZ("Video Quiz"),
  VIEWER_DOWNLOADS("Viewer Downloads"),
  WEBEX_IMPORTING("Webex Importing"),
  ZOOM_IMPORT_FAILED("Zoom Import Failed"),
  ZOOM_IMPORTING("Zoom Importing"),
  ZOOM_VERIFICATION("Zoom Verification"),
  WARNING("WARNING"), // Not documented, but returned by API
}

@Serializable
public data class AuditLogEntry(
  @Serializable(with = LocalDateTimeSerializer::class)
  val creationTimestamp: LocalDateTime? = null,

  val eventName: String = "",

  val details: String = "",
)
