/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity.test

import edu.ucsc.its.temerity.TemClientConfig
import io.github.cdimascio.dotenv.Dotenv
import io.kotest.core.extensions.TestCaseExtension
import io.kotest.core.test.TestCase
import io.kotest.core.test.TestResult
import io.kotest.core.test.TestType
import io.kotest.core.test.isRootTest
import org.koin.core.context.startKoin
import org.koin.core.context.stopKoin
import org.koin.core.module.Module
import org.koin.test.mock.MockProvider
import org.koin.test.mock.Provider

// Utility function from: https://github.com/alvr/katana/blob/d10a53486b95bbc2cfaa0fed636b913957334366/core/tests/src/commonMain/kotlin/dev/alvr/katana/core/tests/KoinExtension.kt
// Originally based on kotest koin extension: https://github.com/kotest/kotest-extensions-koin/blob/30bc2867dd3adfe4e3dc164f97031146563d4435/src/commonMain/kotlin/io.kotest.koin/KoinExtension.kt

fun koinExtension(
  vararg modules: Module,
  mockProvider: Provider<*>? = null,
  mode: KoinLifecycleMode = KoinLifecycleMode.Test,
): TestCaseExtension = KoinExtensionImpl(modules.toList(), mockProvider, mode)

private class KoinExtensionImpl(
  private val modules: List<Module>,
  private val mockProvider: Provider<*>? = null,
  private val mode: KoinLifecycleMode = KoinLifecycleMode.Test,
) : TestCaseExtension {
  @Suppress("TooGenericExceptionCaught")
  override suspend fun intercept(testCase: TestCase, execute: suspend (TestCase) -> TestResult): TestResult = if (testCase.isApplicable()) {
    try {
      stopKoin()
      startKoin {
        if (mockProvider != null) MockProvider.register(mockProvider)
        modules(modules)
      }
      execute(testCase)
    } catch (t: Throwable) {
      throw IllegalStateException(t)
    } finally {
      stopKoin()
    }
  } else {
    execute(testCase)
  }

  private fun TestCase.isApplicable() =
    mode == KoinLifecycleMode.Root &&
      isRootTest() ||
      mode == KoinLifecycleMode.Test &&
      type == TestType.Test
}

enum class KoinLifecycleMode {
  Root,
  Test,
}

internal fun TemClientConfig.configureDevEnvironment(dotenv: Dotenv) {
  serviceUrl = dotenv["YUJADEV_API_URL"]
  serviceToken = dotenv["YUJADEV_TOKEN"]
  optDebugEnabled = true
}

internal fun TemClientConfig.configureProdEnvironment(dotenv: Dotenv) {
  serviceUrl = dotenv["YUJAPROD_API_URL"]
  serviceToken = dotenv["YUJAPROD_TOKEN"]
  optDebugEnabled = true
}
